import filecmp
import pathlib
import shutil
import tempfile
import typing
import zipfile

from terrable import _definitions
from terrable import _s3


def _bundle(
    source_directory: pathlib.Path,
    temp_bundle_directory: pathlib.Path,
) -> pathlib.Path:
    """
    Creates a bundle for the specified module in the temporary bundle directory
    and returns the path to the created zip bundle.
    """
    path = temp_bundle_directory.joinpath(f"{source_directory.name}.zip")

    paths = source_directory.rglob("**/*")
    with zipfile.ZipFile(path, mode="w") as zipper:
        for p in paths:
            zipper.write(p, arcname=p.relative_to(source_directory))

    return path


def _compare(
    context: "_definitions.Context",
    bundle_path: pathlib.Path,
    remote_version: "_definitions.ModuleVersion",
) -> bool:
    """
    Compares the bundled module with the specified remote version and
    returns True if they appear to be identical.
    """
    download_path = bundle_path.parent.joinpath(f"{bundle_path.name}.compare")
    _s3.get_bundle(context, remote_version.key, download_path)
    return filecmp.cmp(str(bundle_path), str(download_path), shallow=False)


def run(context: "_definitions.Context") -> "_definitions.CommandResult":
    """Executes a publish action for the given command context."""
    root_directory = pathlib.Path(context.args.directory).expanduser().absolute()
    module_filters: typing.List[str] = context.args.module_targets
    source_directories = [
        p
        for p in root_directory.iterdir()
        if p.is_dir() and (not module_filters or p.name in module_filters)
    ]
    temp_directory = pathlib.Path(tempfile.mkdtemp())

    for directory in source_directories:
        module_name: str = directory.name
        print(f"\nBUNDLING: {module_name}")

        bundle_path = _bundle(directory, temp_directory)
        print(f"   + Bundled to local path {bundle_path}")

        versions = _s3.get_versions(context, module_name)

        should_publish = (
            context.args.force
            or not versions
            or not _compare(context, bundle_path, versions[-1])
        )
        if not should_publish:
            print(f'   + No changes found. Aborted publishing "{module_name}".')
            continue

        next_version = 1 if not versions else (1 + versions[-1].version)
        print(f"   + Publishing version {next_version}")

        _s3.put_bundle(context, bundle_path, module_name, next_version)
        print(f"   + Module {module_name} has been published")

    shutil.rmtree(temp_directory)

    return _definitions.CommandResult(
        code="PUBLISHED",
        message="Modified module targets have been published.",
    )



import argparse
import dataclasses
import datetime

import boto3


@dataclasses.dataclass(frozen=True)
class Context:
    """Data structure for execution contexts."""

    args: argparse.Namespace
    session: boto3.Session


@dataclasses.dataclass(frozen=True)
class CommandResult:
    """Data structure for command execution results."""

    code: str
    message: str
    data: dict = dataclasses.field(default_factory=lambda: {})


@dataclasses.dataclass(frozen=True)
class ModuleVersion:
    """Data structure for metadata about a bundled module version."""

    #: Name of the module that this is a version of.
    name: str
    #: Response entry from an s3.list_objects_v2 entry response.
    raw: dict

    @property
    def version(self) -> int:
        return int(self.raw.get("Key", "").rsplit("/", 1)[-1].split(".")[0] or "0")

    @property
    def key(self) -> str:
        return self.raw.get("Key", "")

    @property
    def last_modified(self) -> datetime.datetime:
        return self.raw.get("LastModified") or datetime.datetime.utcnow().astimezone(
            datetime.timezone.utc
        )

    @property
    def size(self) -> int:
        """Size of the object in bytes."""
        return self.raw.get("Size") or 0

    def echo(self) -> str:
        """Returns a human-friendly representation of this version."""
        modified = self.last_modified.isoformat()
        return f"{self.version}: {modified} ({self.size:,.0f} bytes)"



import pathlib

import lobotomy

import terrable

MY_DIRECTORY = pathlib.Path(__file__).parent.absolute()
MODULES_DIRECTORY = MY_DIRECTORY.joinpath("modules")


@lobotomy.Patch(path=MY_DIRECTORY.joinpath("test_list.yaml"))
def test_list(lobotomized: lobotomy.Lobotomy):
    """Should execute the list command successfully."""
    terrable.main(["list", "foo-module" "--profile=foo", "--bucket=foo"])



from terrable import _definitions
from terrable import _s3


def run(context: "_definitions.Context") -> "_definitions.CommandResult":
    """Lists version information for the specified command invocation."""
    versions = _s3.get_versions(context, context.args.module_target)

    print(f"\n\n{context.args.module_target}")
    for v in versions:
        print(f"  - {v.echo()}")

    return _definitions.CommandResult(
        code="LISTED",
        message="Modules have been listed.",
    )



import pathlib
from unittest.mock import MagicMock
from unittest.mock import patch

import lobotomy

import terrable

MY_DIRECTORY = pathlib.Path(__file__).parent.absolute()
MODULES_DIRECTORY = MY_DIRECTORY.joinpath("modules")


@patch("filecmp.cmp")
@lobotomy.Patch(path=MY_DIRECTORY.joinpath("test_publish.yaml"))
def test_publish(lobotomized: lobotomy.Lobotomy, filecmp_cmp: MagicMock):
    """Should execute the publish command successfully."""
    lobotomized.add_call("s3", "upload_file", {})
    filecmp_cmp.return_value = False
    terrable.main(["publish", str(MODULES_DIRECTORY), "--profile=foo", "--bucket=foo"])


@patch("filecmp.cmp")
@lobotomy.Patch(path=MY_DIRECTORY.joinpath("test_publish.yaml"))
def test_publish_dry_run(lobotomized: lobotomy.Lobotomy, filecmp_cmp: MagicMock):
    """Should execute the publish command successfully as a dry run operation."""
    filecmp_cmp.return_value = False
    terrable.main(
        [
            "publish",
            str(MODULES_DIRECTORY),
            "--profile=foo",
            "--bucket=foo",
            "--dry-run",
        ]
    )


@patch("filecmp.cmp")
@lobotomy.Patch(path=MY_DIRECTORY.joinpath("test_publish.yaml"))
def test_publish_no_change(lobotomized: lobotomy.Lobotomy, filecmp_cmp: MagicMock):
    """Should execute the publish without uploading an unchanged module."""
    filecmp_cmp.return_value = True
    terrable.main(
        [
            "publish",
            str(MODULES_DIRECTORY),
            "--profile=foo",
            "--bucket=foo",
        ]
    )



import argparse
import typing

import boto3

from terrable import _definitions
from terrable import _lister
from terrable import _publisher


def _parse(arguments: typing.List[str] = None):
    """Parses command line arguments for the publish invocation."""
    parser = argparse.ArgumentParser(allow_abbrev=False)
    subparsers = parser.add_subparsers(dest="command")

    list_parser = subparsers.add_parser("list")
    list_parser.add_argument("module_target")

    publish_parser = subparsers.add_parser("publish")
    publish_parser.add_argument("directory", default=".")
    publish_parser.add_argument("--target", dest="module_targets", action="append")
    publish_parser.add_argument("--dry-run", action="store_true")
    publish_parser.add_argument("--force", action="store_true")

    for p in [list_parser, publish_parser]:
        p.add_argument("--aws-directory", default="~/.aws")
        p.add_argument("--profile", dest="aws_profile")
        p.add_argument("--bucket")
        p.add_argument("--prefix", default="")

    return parser.parse_args(arguments)


def main(arguments: typing.List[str] = None) -> "_definitions.CommandResult":
    """Executes the publish action to deploy the resources to S3."""
    args = _parse(arguments)
    session = boto3.Session(profile_name=args.aws_profile)
    context = _definitions.Context(args, session)

    actions = {
        "publish": _publisher.run,
        "list": _lister.run,
    }
    result = actions[args.command](context)
    print(f"\n\n{result.message}\n\n")
    return result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['terrable', 'terrable.tests']

package_data = \
{'': ['*'], 'terrable.tests': ['modules/foo/*']}

install_requires = \
['boto3>=1.16.16,<2.0.0']

entry_points = \
{'console_scripts': ['terrable = terrable:main']}

setup_kwargs = {
    'name': 'terrable',
    'version': '0.1.0',
    'description': 'Manage private shared terraform modules deployed to S3.',
    'long_description': '# Terrable\n\n[![PyPI version](https://badge.fury.io/py/terrable.svg)](https://pypi.org/project/terrable/)\n[![build status](https://gitlab.com/rocket-boosters/terrable/badges/main/pipeline.svg)](https://gitlab.com/rocket-boosters/terrable/commits/main)\n[![coverage report](https://gitlab.com/rocket-boosters/terrable/badges/main/coverage.svg)](https://gitlab.com/rocket-boosters/terrable/commits/main)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Code style: flake8](https://img.shields.io/badge/code%20style-flake8-white)](https://gitlab.com/pycqa/flake8)\n[![Code style: mypy](https://img.shields.io/badge/code%20style-mypy-white)](http://mypy-lang.org/)\n[![PyPI - License](https://img.shields.io/pypi/l/terrable)](https://pypi.org/project/terrable/)\n\nTerraform private module manager that uses S3 as a backend. Includes simple versioning\nof modules to make forward migration easier. Terraform supports referencing modules\nstored in S3 as compressed files (see\n[S3 Bucket](https://www.terraform.io/docs/modules/sources.html#s3-bucket)\nfor more details). However, managing those packages is not part of Terraform itself.\nThat\'s where *terr&#8226;able* comes in. The terrable CLI allows for bundling terraform\nmodule directories into compressed files and deploying them to S3 with simple\nincremental versioning. That way modules changes can be gradually introduced in\ndependent projects as needed without causing conflicts.\n\n## Installation\n\nTerrable is available via pip:\n\n```shell script\n$ pip install terrable\n```\n\nor via poetry:\n\n```shell script\n# poetry install terrable --dev\n```\n\nOnce installed, the terrable CLI command will be available in your terminal.\n\n## Usage\n\nTerrable operates primarily on a directory that contains within it one or more module\ndirectories. For example:\n\n```\n+---modules\n|   \\---aws-lambda-function\n|           main.tf\n|           output.tf\n|           variables.tf\n|           policy.json\n|\n|   \\---aws-dynamo-db-table\n|           main.tf\n|           output.tf\n|           variables.tf\n```\n\nHere the root "modules" folder contains two modules "aws-lambda-function"\nand "aws-dynamo-db-table". To deploy these as modules via terrable to an S3 bucket\nexecute the command from the parent directory of the modules folder:\n\n```shell script\n$ terrable publish ./modules/ --bucket=<BUCKET_NAME> --profile=<AWS_PROFILE_NAME>\n```\n\nThis command will iterate through each folder inside the modules directory and publish\nany that have changed since their previous publish. Any modules found not to have\nchanged will be skipped. This can be overridden with the `--force` flag. It\'s also\npossible to publish only specific modules within that folder by including the \n`--target=aws-lambda-function` flag. This flag can be specified multiple times to\npublish a select number of specific modules for a given command.\n\nTo inspect modules, there is a list command:\n\n```\n$ terrable list <MODULE_NAME> --bucket=<BUCKET_NAME> --profile=<AWS_PROFILE_NAME>\n```\n\nThis command will print all of the versions and associated metadata for the specified\nmodule.\n',
    'author': 'Scott Ernst',
    'author_email': 'swernst@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/rocket-boosters/terrable',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import typing
import pathlib

from terrable import _definitions


def get_versions(
    context: "_definitions.Context",
    module_name: str,
) -> typing.List["_definitions.ModuleVersion"]:
    """
    Fetches version information for all deployed versions of a given module.
    The versions are sorted from oldest to newest.
    """
    client = context.session.client("s3")
    paginator = client.get_paginator("list_objects_v2")
    kwargs = dict(
        Bucket=context.args.bucket,
        Prefix=f"{context.args.prefix}/{module_name}/",
    )
    results = [
        _definitions.ModuleVersion(module_name, item)
        for page in paginator.paginate(**kwargs)
        for item in page.get("Contents", [])
    ]
    return list(sorted(results, key=lambda s: s.version))


def put_bundle(
    context: "_definitions.Context",
    bundle_path: pathlib.Path,
    module_name: str,
    version: int,
) -> dict:
    """Publishes the version of the module to S3."""
    key = f"{context.args.prefix}/{module_name}/{version}.zip"
    if context.args.dry_run:
        print(f"   ! DRY RUN skipped publishing bundle to {key}")
    else:
        context.session.client("s3").upload_file(
            Filename=str(bundle_path),
            Bucket=context.args.bucket,
            Key=f"{context.args.prefix}/{module_name}/{version}.zip",
            Callback=lambda p: print(f"   + Uploading {module_name} {p:,.0f} bytes"),
            ExtraArgs=dict(
                ContentType="application/zip",
                Metadata={
                    "version": str(version),
                    "module": module_name,
                },
            ),
        )
    return {"key": key}


def get_bundle(
    context: "_definitions.Context",
    key: str,
    download_path: pathlib.Path,
):
    """Downloads the bundle to the specified location."""
    context.session.client("s3").download_file(
        Bucket=context.args.bucket,
        Key=key,
        Filename=str(download_path),
        Callback=lambda p: print(f"   + Downloading {key} {p:,.0f} bytes"),
    )
